//-----------------------------------------------------------------------------
/// usage ./overlayTest -port 505 -visual 0x23
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <X11/extensions/Xvlib.h>

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
const static unsigned long TH_PERIOD = 25;
const static int c_cycleCount = (1000 / TH_PERIOD);
const static Dimension c_width = 800;
const static Dimension c_height = 800;
const static int c_dataSize = 960000;
const static int XvGid = 0x30323449;

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
struct _masks
{
  unsigned int red_mask;
  unsigned int green_mask;
  unsigned int blue_mask;
};

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
Widget UxTopLevel;
XtAppContext UxAppContext;
static bool m_looping;
static Display *Xdisplay;
static Window vWindow;
static Window overlayWindow;
static Pixmap overlayPixmap;
static GC gc;
static GC vgc;
static XvImage *xv_image;
static char* imageData;
static struct _masks colorMask;
static int XvPort = 393; // xvinfo
static int visualId = 0x23; //xdpyinfo

//-----------------------------------------------------------------------------
//// event callback to draw to the drawables
//-----------------------------------------------------------------------------
static void timerCB(XtPointer cd, XtIntervalId *id)
{
  static int count = 0;

  if (count == 8)
  {
    static Pixel color = 0x000005;
    XSetForeground(Xdisplay, gc, color);

    // update the colors for the next overlay update
    Pixel nColor = ((color & colorMask.red_mask) + (0x111111 & colorMask.red_mask)) |
                   ((color & colorMask.green_mask) + (0x222222 & colorMask.green_mask)) |
                   ((color & colorMask.blue_mask) + (0x030303 & colorMask.blue_mask));
    color = nColor;
    if (color >= (colorMask.red_mask | colorMask.green_mask | colorMask.blue_mask))
    {
      color = 0x000002;
    }

    // draw to the overlay pixmap
    XFillRectangle(Xdisplay, overlayPixmap, gc, 500, 500, 300, 300);

    // test known color for 555 to 565 error
    XSetForeground(Xdisplay, gc, 0x7c00 | 0x3e0 | 0x1f); // create white 555 pixel
    XFillRectangle(Xdisplay, overlayPixmap, gc, 50, 50, 300, 300);

    XCopyArea(Xdisplay, overlayPixmap, overlayWindow, gc, 0,0, c_width -100, c_height -100, 0,0);
  }
  else if ((count & 0x1) == 1)
  {
    // update underlay
    for (size_t i = 0; i < (c_dataSize / sizeof(int)); i++)
    {
      // create random background
      const int data = rand();
      ((int *)imageData)[i] = data;
    }

    // push the image to xv
    const int ret = XvPutImage(Xdisplay, XvPort, vWindow, vgc, xv_image,
        0, 0, c_width, c_height, 0, 0, c_width, c_height);

    if (ret < 0)
    {
      fprintf(stderr, "put image error: %d\n", ret);
    }
  }

  // handle reset of counter
  if (++count == c_cycleCount)
  {
    count = 0;
  }

  // re-register the timer callback
  XtAppAddTimeOut(UxAppContext, TH_PERIOD, (XtTimerCallbackProc)timerCB, (XtPointer)NULL);
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
static void mainLoop(XtAppContext app)
{
  XEvent event;
  m_looping = true;
  while (m_looping == true)
  {
    XtAppNextEvent(app, &event);
	if (event.type == DestroyNotify)
	{
	  m_looping = false;
	}
    XtDispatchEvent(&event);
  }
  XSync(XtDisplay(UxTopLevel), True);
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
static int createXvImage(Display *disp, int port, int gid, int width, int height)
{
  imageData = (char *)malloc(c_dataSize);
  if (imageData)
  {
    xv_image = XvCreateImage(disp, port, gid, imageData, width, height);
    if (xv_image == NULL)
    {
      fprintf(stderr, "failed to create the XV image data\n");
      free(imageData);
      imageData = NULL;
      return -1;
    }
  }

  return 0;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
static int createXApp(int argc, char *argv[])
{
  UxTopLevel = XtVaAppInitialize(&UxAppContext, "overlayTest", NULL, 0, &argc, argv, NULL, NULL);
  Xdisplay = XtDisplay(UxTopLevel);
  Widget shell = XtVaAppCreateShell("testScrn", "testShell", applicationShellWidgetClass, Xdisplay, NULL);

  XtInitializeWidgetClass(topLevelShellWidgetClass);
  Widget topShell = XtVaCreatePopupShell("test", topLevelShellWidgetClass, shell, NULL);

  Widget canvas = XtVaCreateManagedWidget("canvas",
    xmDrawingAreaWidgetClass,  topShell,
    XmNresizePolicy,           XmRESIZE_GROW,
    XmNx,                      0,
    XmNy,                      0,
    XmNwidth,                  c_width,
    XmNheight,                 c_height,
    XmNbackground,             0x0000FF,
    NULL);

  // display the app
  XtPopup(topShell, XtGrabNone);

  // get underlay drawable
  vWindow = XtWindow(canvas);
  Screen *screen = XtScreen(canvas);

  // find the 16bit overlay visual
  XVisualInfo vinfoQuery;
  XVisualInfo *vinfo;
  int nitems;

  vinfoQuery.visualid = visualId;
  vinfo = XGetVisualInfo(Xdisplay, VisualIDMask, &vinfoQuery, &nitems);
  if (nitems == 1)
  {
    fprintf(stderr, "visuals found: %d\n", nitems);
    fprintf(stderr, "depth: %d bits_per_rgb %d\n", vinfo->depth, vinfo->bits_per_rgb);
  }
  else
  {
    fprintf(stderr, "no visual found with the id 0x%x\n", visualId);
    return -1;
  }

  // find the overlay window
  struct _overlayInfo
  {
    unsigned long overlay_visual;
    long transparent_type;
    unsigned long value;
    long layer;
  } *overlayInfo = NULL;

  // get overlay atom
  Atom visualAtom = XInternAtom(Xdisplay, "SERVER_OVERLAY_VISUALS", True);
  if (visualAtom != None)
  {
    int actualFormat;
    unsigned long sizeData;
    unsigned long bytesLeft;
    Atom actualType;
    Status status;
    Window root = RootWindow(Xdisplay, vinfo->screen);
    status = XGetWindowProperty(Xdisplay, root, visualAtom, 0L, (long)10000, False, visualAtom, &actualType,
        &actualFormat, &sizeData, &bytesLeft, (unsigned char **)&overlayInfo);

    fprintf(stderr, "status: %d, size: %lu, format: %d\n", status, sizeData, actualFormat);
    fprintf(stderr, "is correct atom: %d\n", actualType == visualAtom);
  }

  // setup a fallback incase overlays are not supported
  if (overlayInfo == NULL)
  {
    overlayInfo = (struct _overlayInfo *)malloc(sizeof(struct _overlayInfo));
    overlayInfo->value = 0xdeadbeef;
    overlayInfo->overlay_visual = 0;
  }

  fprintf(stderr, "transparent color: 0x%lx, visual id 0x%lx\n", overlayInfo->value, overlayInfo->overlay_visual);

  // create the overlay window color map
  Colormap cmap = XCreateColormap(Xdisplay, RootWindowOfScreen(screen), vinfo->visual, AllocNone);

  // create overlay window linked with 16 bit RGB visual
  XSetWindowAttributes winattrs;
  winattrs.colormap = cmap;
  winattrs.background_pixel = overlayInfo->value;
  winattrs.border_pixel = overlayInfo->value;

  overlayWindow = XCreateWindow(Xdisplay, vWindow, 0, 0, c_width -100, c_height -100, 0, vinfo->depth,
      InputOutput, vinfo->visual, CWBackPixel | CWBorderPixel | CWColormap, &winattrs);

  XMapSubwindows(Xdisplay, overlayWindow);
  XMapWindow(Xdisplay, overlayWindow);
  XClearWindow(Xdisplay, overlayWindow);

  // create Graphics context for under/overlay
  XGCValues values;
  gc = XCreateGC(Xdisplay, overlayWindow, 0, &values);
  vgc = XCreateGC(Xdisplay, vWindow, 0, &values);

  // create XvImage
  createXvImage(Xdisplay, XvPort, XvGid, c_width, c_height);

  // set underlay foreground to white
  XSetForeground(Xdisplay, vgc, 0xFFFFFF);

  // verify green mask
  fprintf(stderr, "green 0x%lx\n", vinfo->green_mask);

  colorMask.red_mask = vinfo->red_mask;
  colorMask.green_mask = vinfo->green_mask;
  colorMask.blue_mask = vinfo->blue_mask;

  // set the overlay foreground
  XSetForeground(Xdisplay, gc, 0);
  XFillRectangle(Xdisplay, overlayWindow, gc, 0,0, c_width -100, c_height -100);

  // create the overlay pixmap
  overlayPixmap = XCreatePixmap(Xdisplay, overlayWindow, c_width -100, c_height -100, vinfo->depth);

  // clear both the overlay window and the pixmap
  if (vinfo->depth == 16)
  {                             // 0x7fff
    //XSetForeground(Xdisplay, gc, 0x8000); // red, confirms its 565
    XSetForeground(Xdisplay, gc, 0xF0000); // black
  }
  XFillRectangle(Xdisplay, overlayPixmap, gc, 0,0, c_width -100, c_height -100);

  // register the timer callback
  XtAppAddTimeOut(UxAppContext, TH_PERIOD, (XtTimerCallbackProc)timerCB, (XtPointer)NULL);

  // clean up resources
  XFree(vinfo);
  free(overlayInfo);

  // good exit status
  return 0;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
  // parse args
  if (argc > 1)
  {
    if (!((argc == 3) || (argc == 5)))
    {
      fprintf(stderr, "not the right number of options %d\n", argc -1);
      exit(1);
    }

    int opt = 0;
    for (int i = 1; i < argc; i+=2)
    {
      if (strcmp(argv[i], "-port") == 0)
      {
        if ((opt & 0x1) == 0)
        {
          int port;
          opt |= 0x1;
          int ret = sscanf(argv[i+1], "%d", &port);
          if (ret == 1)
          {
            XvPort = port;
          }
          else
          {
            fprintf(stderr, "bad option for port\n");
            exit(1);
          }
        }
        else
        {
          fprintf(stderr, "port specified twice\n");
          exit(1);
        }
      }
      else if (strcmp(argv[i], "-visual") == 0)
      {
        if ((opt & 0x2) == 0)
        {
          opt |= 0x2;
          int vis;
          const int ret = sscanf(argv[i+1], "0x%x", &vis);
          if (ret == 1)
          {
            visualId = vis;
          }
        }
        else
        {
          fprintf(stderr, "visual specified twice\n");
          exit(1);
        }
      }
      else
      {
        fprintf(stderr, "unknown option |%s|\n", argv[2]);
        exit(1);
      }
    }
  }

  // do what we are here for
  const int ret = createXApp(argc, argv);
  if (ret == 0)
  {
    // start the X process events loops
    mainLoop(UxAppContext);
  }

  // after loop exit
  // TODO memory leaks
  exit(0);
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
